/*
 * $Id: liblzostream.cpp,v 1.2 2014-02-10 10:28:29 swaj Exp $
 * Was written by Alexey Semenoff (c) 2012-2014
 * See COPYING for copyiright policy
 */

#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <lzo/lzo1x.h>
#include <algorithm>
#include "liblzostream.h"

//------------------------------------------------------------------------------
unsigned int CLzoStream::stream_PAGE_SIZE = (8*1024);
const unsigned int CLzoStream::const_PAGE_SIZE_MIN = (256);
const unsigned int CLzoStream::const_PAGE_SIZE_MAX = (128*1024);
const int CLzoStream::const_CRC_INITIAL = 1; 
const unsigned int CLzoStream::const_SIGLEN = sizeof (uint8_t);
const unsigned int CLzoStream::const_HEADER_LEN = (3*sizeof(uint32_t) + const_SIGLEN);
const unsigned int CLzoStream::const_INT_BUF_SIZE = (2*const_PAGE_SIZE_MAX);
const uint8_t CLzoStream::const_ESigLzo1xAdler = ESigLzo1xAdler;

//------------------------------------------------------------------------------
/* static */
int CLzoStream::lzoSetPageSize(unsigned int psize)
{
    if (psize >= const_PAGE_SIZE_MIN && psize <= const_PAGE_SIZE_MAX)
    {
        stream_PAGE_SIZE = psize;
        return EOk;
    }
    return ENotSupported;
}

//------------------------------------------------------------------------------
/* static */
int CLzoStream::lzoSetPageSize(lzo_stream *s, unsigned int psize)
{
    if (s->total_in || s->total_out) { // not allowed to change in fly
        return ENotSupported;
    }

    if (psize >= const_PAGE_SIZE_MIN && psize <= const_PAGE_SIZE_MAX)
    {
        ((internal_state*)s->int_st)->set_page_size(psize);
        return EOk;
    }

    return ENotSupported;
}

//------------------------------------------------------------------------------
/* static */
int CLzoStream::lzoGetPageSize(lzo_stream *s)
{
    return ((internal_state*)s->int_st)->get_page_size();
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoInflateReset(lzo_stream *s)
{
    s->reset_stat();
    ((internal_state*)s->int_st)->reset();
    return EOk;
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoFlushIntCache(lzo_stream *s)
{
    ((internal_state*)s->int_st)->reset();
    return EOk;
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoDeflateReset(lzo_stream *s)
{
    return CLzoStream::lzoInflateReset(s);
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoDeflateInit(lzo_stream *s)
{
    lzo_init();
    s->reset_stat();

    voidp int_stp = malloc(sizeof(internal_state));

    if (! int_stp) {
        return EMemError;
    }
    memset (int_stp, 0x0, sizeof(internal_state) );

    voidp in_bufp = malloc(const_INT_BUF_SIZE);
    if (! in_bufp) {
        free(int_stp);
        return EMemError;
    }

    voidp out_bufp = malloc(const_INT_BUF_SIZE);
    if (! out_bufp) {
        free(int_stp);
        free(in_bufp);
        return EMemError;
    }

    voidp wrkmem = malloc(LZO1X_1_MEM_COMPRESS);
    if (! wrkmem ) {
        free(int_stp);
        free(in_bufp);
        free(out_bufp);
        return EMemError;
    }

    internal_state *st = (internal_state*) int_stp;
    st->in_buffer = (uint8_t*) in_bufp;
    st->out_buffer = (uint8_t*) out_bufp;
    st->wrkmem = wrkmem;
    st->page_size = stream_PAGE_SIZE;
    
    s->int_st = int_stp;
    return EOk;
}
//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoInflateInit(lzo_stream *s)
{
    return  lzoDeflateInit(s);
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoDeflateEnd(lzo_stream *s)
{
    internal_state *st = (internal_state*)s->int_st;

    if (st) {
        if (st->in_buffer) {
            free(st->in_buffer);
            st->in_buffer = 0;
        }

        if (st->out_buffer) {
            free(st->out_buffer);
            st->out_buffer = 0;
        }

        if (st->wrkmem) {
            free(st->wrkmem);
            st->wrkmem=0;
        }
        free(st);
    }

    s->int_st = 0;
    s->reset();

    return EOk;
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoInflateEnd(lzo_stream *s)
{
    return lzoDeflateEnd(s);
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::pack_chunk_from_stream (lzo_stream *s, size_t bytes_to_pack)
{
    internal_state *st = (internal_state*)s->int_st;
    uint32_t crc = lzo_adler32 (const_CRC_INITIAL, (lzo_bytep)s->next_in, bytes_to_pack);
    size_t dest_sz = const_INT_BUF_SIZE;
    uint8_t *bff;
    
    int zip_rv = lzo1x_1_compress ( (lzo_bytep)s->next_in, bytes_to_pack,
                                    (st->out_buffer + const_HEADER_LEN), (lzo_uint*) &dest_sz,
                                    st->wrkmem );
    if (zip_rv) {
        return getError(ELzo, zip_rv);
    }

    bff = st->out_buffer; 
    *(bff) = const_ESigLzo1xAdler;
    bff++;
    *((uint32_t*)(bff)) = (uint32_t)dest_sz;
    bff +=4;
    *((uint32_t*)(bff)) = (uint32_t)bytes_to_pack;
    bff +=4;
    *((uint32_t*)(bff)) = crc;

    size_t to_memcpy=0;
    int rc = EOk;

    if ( (dest_sz + const_HEADER_LEN) > st->av_out ) {
        to_memcpy = st->av_out;
        st->av_out_cached = dest_sz + const_HEADER_LEN - st->av_out;
        rc=EOutputBufFull;
    }
    else {
        to_memcpy = dest_sz + const_HEADER_LEN;
    }
    
    memcpy((voidp)s->next_out, (voidp)st->out_buffer, to_memcpy );

    if (rc==EOutputBufFull) {
        memmove((voidp)st->out_buffer, (voidp)(st->out_buffer+to_memcpy), st->av_out_cached);  
    }

    s->next_out  += to_memcpy;
    s->total_out += to_memcpy;
    s->avail_out += to_memcpy;
    st->av_out   -= to_memcpy;

    s->next_in  += bytes_to_pack; 
    s->total_in += bytes_to_pack;
    s->avail_in -= bytes_to_pack;

    return rc;
}

//------------------------------------------------------------------------------
/*static*/ 
int CLzoStream::pack_chunk_from_cache ( lzo_stream *s, size_t bytes_to_pack, size_t bytes_to_report)
{
    internal_state *st = (internal_state*)s->int_st;
    size_t dest_sz = const_INT_BUF_SIZE;
    uint32_t crc = lzo_adler32 (const_CRC_INITIAL, st->in_buffer,  bytes_to_pack );
    uint8_t *bff;

    int zip_rv = lzo1x_1_compress ( (lzo_bytep)st->in_buffer,  bytes_to_pack,
                                    st->out_buffer + const_HEADER_LEN, (lzo_uint*) &dest_sz,
                                    st->wrkmem );
    if (zip_rv) {
        return getError(ELzo, zip_rv);
    }

    bff = st->out_buffer; 
    *(bff) = const_ESigLzo1xAdler;
    bff++;
    *((uint32_t*)(bff)) = (uint32_t)dest_sz;
    bff +=4;
    *((uint32_t*)(bff)) = (uint32_t)bytes_to_pack;
    bff +=4;
    *((uint32_t*)(bff)) = crc;
    
    size_t to_memcpy=0;
    int rc = EOk;

    if ( (dest_sz + const_HEADER_LEN) > st->av_out ) {
        to_memcpy = st->av_out;
        st->av_out_cached = dest_sz + const_HEADER_LEN - st->av_out;
        rc=EOutputBufFull;
    }
    else {
        to_memcpy = dest_sz + const_HEADER_LEN;
    }

    memcpy((voidp)s->next_out, (voidp)st->out_buffer, to_memcpy);

    if (rc==EOutputBufFull) {
        memmove((voidp)st->out_buffer, (voidp)(st->out_buffer+to_memcpy), st->av_out_cached);  
    }

    s->next_out  += to_memcpy;
    s->total_out += to_memcpy;
    s->avail_out += to_memcpy;
    st->av_out   -= to_memcpy;

    s->next_in  += bytes_to_report; 
    s->total_in += bytes_to_report;
    s->avail_in -= bytes_to_report;

    return rc;
}

//------------------------------------------------------------------------------
/*static*/ 
int CLzoStream::pack_chunks_flush( lzo_stream *s)
{
    internal_state *st = (internal_state*)s->int_st;
    uint full_chunks = s->avail_in / st->page_size;
    uint chunk_tail = s->avail_in % st->page_size;
    int zip_rv=0;

    for (uint bl=0; bl<full_chunks; bl++) { // full chunks first
        zip_rv = pack_chunk_from_stream (s, st->page_size );
                
        if (zip_rv == EOutputBufFull) {
            return EOk;
        }
        if (zip_rv) {
            return getError(ELzo, zip_rv);
        }
    }
    if (chunk_tail) { // pack chunk tail
        zip_rv = pack_chunk_from_stream (s, chunk_tail);
        
        if (zip_rv == EOutputBufFull) {
            return EOk;
        }
        
        if (zip_rv) {
            return getError(ELzo, zip_rv);
        }
    }
    return EStreamEnd;
}

//------------------------------------------------------------------------------
/*static*/ 
int CLzoStream::pack_chunks_noflush( lzo_stream *s)
{
    internal_state *st = (internal_state*)s->int_st;
    uint full_chunks = s->avail_in / st->page_size;
    uint chunk_tail = s->avail_in % st->page_size;
    int zip_rv=0;
    
    for (uint bl=0; bl<full_chunks; bl++) { // full chunks first
        zip_rv = pack_chunk_from_stream (s, st->page_size );

        if (zip_rv == EOutputBufFull) {
            return EOk;
        }
        if (zip_rv) {
            return getError(ELzo, zip_rv);
        }
    }
            
    // store tail if any
    if (chunk_tail)
    {
        memcpy ( (voidp)st->in_buffer, (voidp)s->next_in, chunk_tail);
        st->av_in = chunk_tail;
        s->next_in  += chunk_tail; 
        s->total_in += chunk_tail;
        s->avail_in -= chunk_tail;
    }
    
    return EOk;
}
//------------------------------------------------------------------------------
/*static*/ 
int CLzoStream::lzoDeflate( lzo_stream *s, int mode)
{
    internal_state *st = (internal_state*)s->int_st;
    int zip_rv=0, rc=EOk;
    uint32_t av_out=0;

    if (! s->avail_out ) {
        return EBufError;
    }

    prepare_stream(s, st, av_out);

    rc = EOk;
    do {
        if (st->av_out_cached)
        {
            if ( st->av_out < st->av_out_cached ) // not enough to store all data
            {
                memcpy (s->next_out, st->out_buffer, st->av_out);
                s->next_out += st->av_out;
                s->avail_out = st->av_out;
                s->total_out += st->av_out;
                memmove ( st->out_buffer, st->out_buffer + st->av_out, st->av_out_cached - st->av_out );
                st->av_out_cached -= st->av_out;
                break;
            }
            else
            {
                memcpy (s->next_out, st->out_buffer, st->av_out_cached);
                s->next_out += st->av_out_cached;
                s->avail_out = st->av_out_cached;
                s->total_out += st->av_out_cached;
                st->av_out_cached = 0;

                if ( mode != ELzoNoFlush && ! s->avail_in) // flush + nothing to do
                {
                    rc = EStreamEnd;
                }
                break;
            }
        }

        if ( mode != ELzoNoFlush) // flush stream mode
        {
            if (!st->av_in) //  nothing in cache yet, all in should be compressed
            {
                rc = pack_chunks_flush(s);
                break;
            }
            else //\\ something in cache
            {
                if (st->av_in + s->avail_in < st->page_size /*stream_PAGE_SIZE*/ )
                {
                    uint8_t *dest = st->in_buffer + st->av_in;
                    memcpy ((voidp)dest, (voidp)s->next_in, s->avail_in);
                    st->av_in += s->avail_in;
                
                    zip_rv = pack_chunk_from_cache (s, st->av_in, s->avail_in);
                    st->av_in =0; // empty cache
                
                    if (zip_rv == EOutputBufFull) {
                        break;
                    }

                    if (zip_rv) {
                        rc = getError(ELzo, zip_rv);
                        break;
                    }

                    rc = EStreamEnd;
                    break;
                }
                else // (st->av_in + s->avail_in >= PAGE_SIZE
                {
                    uint8_t *dest = st->in_buffer + st->av_in;
                    uint to_copy = st->page_size /*stream_PAGE_SIZE*/ - st->av_in;
                    memcpy ((voidp)dest, (voidp)s->next_in, to_copy);

                    zip_rv = pack_chunk_from_cache (s, st->page_size /*stream_PAGE_SIZE*/, to_copy);
                    st->av_in = 0; // empty cache

                    if (zip_rv == EOutputBufFull) {
                        break;
                    }
                    if (zip_rv) {
                        rc = getError(ELzo, zip_rv);
                        break;
                    }

                    rc = pack_chunks_flush(s);
                    break;
                }

            } //\\ else //\\ something in cache

        } //\\  if ( mode != LZO_NO_FLUSH ) // flush stream mode

        else // mode =  LZO_NO_FLUSH aka "lazy mode"
        {
            if (!st->av_in) //  nothing in cache yet, all in should be compressed
            {
                rc = pack_chunks_noflush(s);
                break;
            }
            else // something in cache already
            {
                if (st->av_in + s->avail_in < st->page_size /*stream_PAGE_SIZE*/)
                {
                    uint8_t *dest = st->in_buffer + st->av_in;
                    memcpy ((voidp)dest, (voidp)s->next_in, s->avail_in);
                    st->av_in += s->avail_in;

                    s->next_in  += s->avail_in;
                    s->total_in += s->avail_in;
                    s->avail_in = 0;

                    break;
                }
                else // st->av_in + s->avail_in >= PAGE_SIZE
                {
                    uint8_t *dest = st->in_buffer + st->av_in;
                    uint to_copy = st->page_size /*stream_PAGE_SIZE*/ - st->av_in;
                    memcpy ((voidp)dest, (voidp)s->next_in, to_copy);
                    zip_rv = pack_chunk_from_cache (s, st->page_size /*stream_PAGE_SIZE*/, to_copy);
                    st->av_in =0; // empty cache

                    if (zip_rv == EOutputBufFull) {
                        break;
                    }

                    if (zip_rv) {
                        rc = getError(ELzo, zip_rv);
                        break;
                    }

                    rc = pack_chunks_noflush(s);
                    break;
                }
            } //\\      else // something in cache already
        } //\\ // mode =  LZO_NO_FLUSH
    } while (0);

    s->update_avail_out (av_out);

    return rc;
}

//------------------------------------------------------------------------------
/*static*/ 
int CLzoStream::lzoInflate(lzo_stream *s, int /*mode*/ )
{
     internal_state *st = (internal_state*)s->int_st;
     int zip_rv=0, rc=EOk;
     uint32_t av_out, comp_size=0, ucomp_size=0, crc=0;
     size_t av_out_sz=0;

     if (! s->avail_out) {
         return EBufError;
     }

     if (! s->avail_in && ! st->av_out_cached ) { // no more data
         return EOk;
     }

     prepare_stream(s, st, av_out);

     if (st->av_out_cached)
     {
         if ( st->av_out < st->av_out_cached ) // not enough for all data
         {
             memcpy (s->next_out, st->out_buffer, st->av_out);
             s->next_out += st->av_out;
             s->avail_out = st->av_out;
             s->total_out +=  st->av_out;
             memmove ( st->out_buffer, st->out_buffer + st->av_out, st->av_out_cached - st->av_out );
             st->av_out_cached -= st->av_out;
             s->update_avail_out (av_out);
             return EOk; 
         }
         else
         {
             memcpy (s->next_out, st->out_buffer, st->av_out_cached);
             s->next_out += st->av_out_cached;
             s->avail_out = st->av_out_cached;
             st->av_out -= st->av_out_cached; 
             s->total_out += st->av_out_cached;
             st->av_out_cached = 0;
         }
     }
     
     rc = EOk;
     while (1)
     {
         if (! s->avail_in) {
             rc = EStreamEnd;
             break;
         }

         if (st->av_in < const_HEADER_LEN)
         {
             uint8_t *dest = st->in_buffer + st->av_in;
             bool header_ok = ( st->av_in + s->avail_in < const_HEADER_LEN ) ? false : true;
             uint to_copy = std::min<uint32_t> ( (const_HEADER_LEN-st->av_in), s->avail_in);
         
             memcpy ((voidp)dest, (voidp)s->next_in, to_copy);

             st->av_in += to_copy;
             s->next_in += to_copy;
             s->total_in += to_copy;
             s->avail_in -= to_copy;

             if (! header_ok) {
                 break;
             }
         }

         if (*st->in_buffer != const_ESigLzo1xAdler)
         {
             rc = EDataError;
             break;
         }

         comp_size = *((uint32_t*)( st->in_buffer + 1));
         ucomp_size = *((uint32_t*)( st->in_buffer + 5));
         crc = *((uint32_t*)( st->in_buffer + 9));
     
         if ( st->av_in + comp_size > const_INT_BUF_SIZE ) // out of buffer
         {
             rc = EMemError;
             break;
         }

         if ( (s->avail_in + st->av_in) < (comp_size + const_HEADER_LEN ) )  // not enough to decompress chunk
         {
             uint8_t *dest = st->in_buffer + st->av_in;
             memcpy ((voidp)dest, (voidp)s->next_in, s->avail_in);
             st->av_in += s->avail_in;
             s->next_in  += s->avail_in;
             s->total_in += s->avail_in;
             s->avail_in -= s->avail_in;
    
             break;
         }

         // copy exact chunk
         {
             uint8_t *dest = st->in_buffer + st->av_in;
             uint to_copy = (comp_size + const_HEADER_LEN) - st->av_in;
    
             memcpy ((voidp)dest, (voidp)s->next_in, to_copy);
             st->av_in   += to_copy;
             s->next_in  += to_copy;
             s->total_in += to_copy;
             s->avail_in -= to_copy;

         }

         if (ucomp_size > st->av_out ) // out buff too small!
         {
             av_out_sz = const_INT_BUF_SIZE;

             zip_rv = lzo1x_decompress_safe
                 ( (st->in_buffer + const_HEADER_LEN), comp_size,
                   st->out_buffer, (lzo_uint*) &av_out_sz,  
                   st->wrkmem  /* NOT USED */ );

             if (zip_rv)
             {
                 rc = getError(ELzo, zip_rv);
                 break;
             }

             if ( (ucomp_size !=  av_out_sz) ||
                  (crc != lzo_adler32 (const_CRC_INITIAL, (lzo_bytep)st->out_buffer, ucomp_size))
                  ) {
                 rc = EStreamError;
                 break;
             }

             size_t to_copy = st->av_out;
             memcpy ( s->next_out, st->out_buffer, to_copy);
             s->avail_out += to_copy;
             s->next_out  += to_copy; 
             s->total_out += to_copy;
             st->av_out = 0;
             st->av_in = 0;
             st->av_out_cached = ucomp_size - to_copy;
             memmove (st->out_buffer, st->out_buffer + to_copy, st->av_out_cached);

             break;

         } //\\ !NOT out buff too small!
         else
         {
             size_t av_out_sz = st->av_out;

             zip_rv = lzo1x_decompress_safe
                 ( (st->in_buffer + const_HEADER_LEN), comp_size,
                   s->next_out, (lzo_uint*) &av_out_sz,  
                   st->wrkmem  /* NOT USED */ );

             if (zip_rv)
             {
                 rc = getError(ELzo, zip_rv);
                 break;
             }

             if ( (ucomp_size !=  av_out_sz) ||
                  (crc != lzo_adler32 (const_CRC_INITIAL, (lzo_bytep)s->next_out, ucomp_size))
                  ) {

                 rc =  EStreamError;
                 break;
             }

             s->avail_out += av_out_sz;
             s->next_out  += av_out_sz; 
             st->av_out   -= av_out_sz;
             s->total_out += av_out_sz;
             st->av_in = 0;

             if (! st->av_out) {
                 break;
             }

         }
     } //\\ while(1)

     s->update_avail_out (av_out);

     return rc;
}

//------------------------------------------------------------------------------
/*static*/
int CLzoStream::lzoInflateCached (lzo_stream *s)
{
    internal_state *st = (internal_state*)s->int_st;
    return st->av_out_cached;
}


//------------------------------------------------------------------------------
/*static*/
const uint8_t* CLzoStream::lzoInflateCachedPtr (lzo_stream *s)
{
    internal_state *st = (internal_state*)s->int_st;
    return (const uint8_t*) st->out_buffer;
}

//------------------------------------------------------------------------------

// __END__
